import redis
from .strings import STRINGS

strings = STRINGS['database']


class Database:

    class NotActiveError(Exception):

        def __init__(self, service, name):
            message = strings['NOT_ACTIVE_ERROR'].format(
                {"service": service, "name": name})
            Exception.__init__(self, message)

    def __init__(self, connection=None):
        if connection is None:
            connection = redis.StrictRedis(host='127.0.0.1', port=6379, db=0)

        self._connection = connection
        self._db = connection

        self._active = True

    def destroy(self):
        self.close()
        self._db = None
        self._connection = None

    def get_connection(self):
        if not self._active:
            raise Database.NotActiveError()
        return self._connection

    def get(self):
        return self._db

    def hash_key_exists(self, hash_key, key):
        return self._db.hexists(hash_key, key)

    def get_hash_key(self, hash_key, key):
        return self._db.hget(hash_key, key)

    def get_all_hash_keys(self, hash_key):
        return self._db.hgetall(hash_key)

    def set_hash_key_if_not_exists(self, hash_key, key, value):
        was_set = self._db.hsetnx(hash_key, key, value)
        if was_set == 1:
            return True
        else:
            return False

    def del_hash_key(self, hash_key, key):
        res = self._db.hdel(hash_key, key)
        if res == 1:
            return True
        else:
            return False

    def del_key(self, key):
        res = self._db.delete(key)
        if res == 1:
            return True
        else:
            return False

    def flush(self):
        return self._db.flushdb()

    def close(self):
        self._active = False
        # Disconnecting will be handled automatically
        # There is no good way to do it manualy because of
        # connection pool :(
        # self._db.connection_pool.disconnect()


class Transaction(Database):

    def __init__(self, connection):
        Database.__init__(self, connection)
        self._db = connection.pipeline()
        self._db.multi()

    def commit(self):
        if self._db is None:
            raise RuntimeError('No transaction in progress')

        self._db.execute()
        self._db = None
        self._connection = None
