import time
import inspect
from shortid import ShortId
from .database import Database, Transaction
from .subscriber import Subscriber
from .publisher import Publisher
from .message import Message
from .pubsub import PubSub
from .strings import STRINGS
from .lib.threadpool import ThreadPool
from .errors import IDCollisionError

rpc_strings = STRINGS['rpc']
expand = [
    'CONSUMMER_NAMESPACE',
    'CONSUMMER_CALL_RESULT_NAMESPACE',
    'RPC_CALL_TYPE',
    'RPC_ERROR_TYPE',
    'RPC_RESULT_TYPE',
    'STORAGE_SERVICE_NAMESPACE',
    'STORAGE_ACTIVE_SERVICES_COLLECTION'
]

for key in expand:
    rpc_strings[key] = rpc_strings[key].format(rpc_strings)

KEY_DELIMITER = rpc_strings['KEY_DELIMITER']
NAMESPACE = rpc_strings['NAMESPACE']
CONSUMMER_NAMESPACE = rpc_strings['CONSUMMER_NAMESPACE']
CONSUMMER_CALL_RESULT_NAMESPACE = rpc_strings[
    'CONSUMMER_CALL_RESULT_NAMESPACE']
RPC_CALL_TYPE = rpc_strings['RPC_CALL_TYPE']
RPC_ERROR_TYPE = rpc_strings['RPC_ERROR_TYPE']
RPC_RESULT_TYPE = rpc_strings['RPC_RESULT_TYPE']
STORAGE_SERVICE_NAMESPACE = rpc_strings['STORAGE_SERVICE_NAMESPACE']
STORAGE_ACTIVE_SERVICES_COLLECTION = rpc_strings[
    'STORAGE_ACTIVE_SERVICES_COLLECTION']
METHODS_COLLECTION = rpc_strings['METHODS_COLLECTION']

thread_pool = ThreadPool()


class Rpc(PubSub):

    class MethodExistsError(Exception):

        def __init__(self, service, name):
            message = rpc_strings['METHOD_EXISTS_ERROR'].format(
                {'service': service, 'name': name})
            Exception.__init__(self, message)

    class MethodDoesNotExistsError(Exception):

        def __init__(self, service, name):
            message = rpc_strings['METHOD_DOES_NOT_EXISTS_ERROR'].format(
                {'service': service, 'name': name})
            Exception.__init__(self, message)

    class RpcMessage(Message):

        def __init__(self, id, *args, **kwargs):
            Message.__init__(self, *args, **kwargs)
            self['id'] = id

    namespace = NAMESPACE

    def __init__(self):
        self._setuped = False

    def is_setuped(self):
        return self._setuped

    def create_key(self, *args):
        return KEY_DELIMITER.join(args)


class RpcService(Rpc):

    class ServiceAlreadyActiveError(Exception):

        def __init__(self, service):
            message = rpc_strings[
                'SERVICE_ALREADY_ACTIVE_ERROR'].format({'service': service})
            Exception.__init__(self, message)

    class ServiceIsNotActiveError(Exception):

        def __init__(self, service):
            message = rpc_strings[
                'SERVICE_IS_NOT_ACTIVE_ERROR'].format({'service': service})
            Exception.__init__(self, message)

    class CallResultMessage(Rpc.RpcMessage):
        pass

    class CallMessage(Rpc.RpcMessage):

        def __init__(self, method, args, *fargs, **kwargs):
            Rpc.RpcMessage.__init__(self, *fargs, **kwargs)
            self['method'] = method
            self['args'] = args

    def __init__(self, service_name):
        Rpc.__init__(self)
        self.name = service_name

        self._db = Database()
        self._methods = {}

        self._pub = Publisher()
        self._sub = Subscriber()
        self._sub.add_message_handler(self._message_handler)

        self.methods_collection_name = self.create_key(
            self.get_service_key(), METHODS_COLLECTION)
        self.service_channel_name = self.get_service_key()

    def is_active(self):
        return self._db.hash_key_exists(STORAGE_ACTIVE_SERVICES_COLLECTION, self.name)

    def get_service_key(self, service_name=None):
        return self.create_key(STORAGE_SERVICE_NAMESPACE, service_name or self.name)

    def method_exists(self, name):
        return self._db.hash_key_exists(self.methods_collection_name, name)

    def _message_handler(self, message):
        thread_pool.add_task(self, '_call_task', message)

    def _call_task(self, message):
        res_type = RPC_RESULT_TYPE
        res = None

        try:
            if not self._setuped:
                raise RpcService.ServiceIsNotActiveError(self.name)

            if message['type'] == RPC_CALL_TYPE:
                res = self._call_method(message['method'], message['args'])
            else:
                raise RuntimeError(
                    'No handler for type `{0}`'.format(message['type']))
        except Exception as err:
            res_type = RPC_ERROR_TYPE
            res = err.message
        finally:
            self._publish_call_result(message, res_type=res_type, res=res)

    '''Calls method in threadpool'''

    def _call_method(self, name, args=()):
        method = self._methods.get(name, None)
        if method is None:
            raise Rpc.MethodDoesNotExistsError(self.name, name)
        return method(*args)

    def _publish_call_result(self, message, res_type=None, res=None):
        reply = RpcService.CallResultMessage(
            id=message['id'],
            type=res_type,
            data=res,
            replyTo=message['channel']
        )

        self._pub.publish(message['replyTo'], reply)

    def setup(self):
        if self._setuped:
            raise RpcService.ServiceAlreadyActiveError(self.name)

        res = self._db.set_hash_key_if_not_exists(
            STORAGE_ACTIVE_SERVICES_COLLECTION, self.name, int(time.time()))
        if not res:
            raise RpcService.ServiceAlreadyActiveError(self.name)

        self._sub.subscribe(self.service_channel_name)
        self._setuped = True

        return True

    def shutdown(self):
        if not self._setuped:
            raise RpcService.ServiceIsNotActiveError(self.name)

        t = Transaction(self._db.get_connection())
        t.del_hash_key(STORAGE_ACTIVE_SERVICES_COLLECTION, self.name)
        t.del_key(self.methods_collection_name)
        t.commit()

        self._sub.unsubscribe(NAMESPACE)
        self._sub.destroy()
        self._setuped = False

        return True

    def define_method(self, name, fn):
        if not self._setuped:
            raise RpcService.ServiceIsNotActiveError(self.name)

        res = self._db.set_hash_key_if_not_exists(
            self.methods_collection_name, name, len(inspect.getargspec(fn).args))
        if not res:
            raise Rpc.MethodExistsError(self.name, name)

        self._methods[name] = fn
        return True

    def get_methods(self):
        res = self._db.get_all_hash_keys(self.methods_collection_name)

        return res


class RpcConsummer(Rpc):

    class ConsummerAlreadyActiveError(Exception):

        def __init__(self, service):
            message = rpc_strings[
                'CONSUMMER_ALREADY_ACTIVE_ERROR'].format({'service': service})
            Exception.__init__(self, message)

    class ConsummerIsNotActiveError(Exception):

        def __init__(self, service):
            message = rpc_strings[
                'CONSUMMER_IS_NOT_ACTIVE_ERROR'].format({'service': service})
            Exception.__init__(self, message)

    class EndpointError(Exception):
        pass

    class QueueItem:

        def __init__(self, id, date, on_result, on_error):
            self.id = id
            self.date = date
            self.on_result = on_result
            self.on_error = on_error
            self._data = {}

        def set(self, key, value):
            self._data[key] = value

        def get(self, key):
            return self._data.get(key)

    def __init__(self, service_name):
        Rpc.__init__(self)
        self.service = RpcService(service_name)

        self._sid = ShortId()
        self._sub = Subscriber()
        self._sub.add_message_handler(self._on_call_result)
        self._pub = Publisher()
        self._pending_calls = {}

    def _enqueue_call(self, on_result, on_error):
        id = self._sid.generate()

        if self._get_queud_call(id) is not None:
            raise IDCollisionError(id)

        item = RpcConsummer.QueueItem(
            id=id,
            date=int(time.time() * 1000),
            on_result=on_result,
            on_error=on_error
        )

        self._pending_calls[id] = item
        return item

    def _dequeue_call(self, id):
        self._pending_calls.pop(id)

    def _get_queud_call(self, id):
        return self._pending_calls.get(id)

    def _pop_queued_call(self, id):
        queue_item = self._get_queud_call(id)
        self._dequeue_call(id)
        return queue_item

    def _on_call_result(self, msg):
        msg_id = msg.get('id')
        queue_item = self._pop_queued_call(msg_id)

        msg_type = msg.get('type')
        if msg_type != RPC_RESULT_TYPE and msg_type != RPC_ERROR_TYPE:
            raise RuntimeError(
                rpc_strings['UNKNOWN_MESSAGE_TYPE_ERROR'].format({'type': msg_type}))

        if queue_item is None:
            raise RuntimeError(
                rpc_strings['NO_SUCH_ITEM_IN_QUEUE_ERROR'].format({'id': msg_id}))

        thread_pool.add_task(self, '_handle_queue_item_task', msg, queue_item)

    def _handle_queue_item_task(self, msg, queue_item):
        self._sub.unsubscribe(queue_item.get('channel'))
        if not self._setuped:
            queue_item.on_error(
                RpcConsummer.ConsummerIsNotActiveError(self.service.name))
            return

        if msg.get('type') == RPC_ERROR_TYPE:
            queue_item.on_error(RpcConsummer.EndpointError(msg.get('data')))
        else:
            queue_item.on_result(msg.get('data'))

    def setup(self):
        if self._setuped:
            raise RpcConsummer.ConsummerAlreadyActiveError(
                self.service.name)

        self._pending_calls = {}

        if not self.service.is_active():
            raise RpcService.ServiceIsNotActiveError(self.service.name)

        self._setuped = True
        return True

    def shutdown(self):
        if not self._setuped:
            raise RpcConsummer.ConsummerIsNotActiveError(self.service.name)

        self._pending_calls = {}
        self._setuped = False
        return True

    def call(self, method_name, args, on_result, on_error):
        queue_item = self._enqueue_call(on_result, on_error)
        id = queue_item.id
        try:
            reply_to = self.create_key(CONSUMMER_CALL_RESULT_NAMESPACE, id)
            msg = RpcService.CallMessage(
                id=id,
                type=RPC_CALL_TYPE,
                method=method_name,
                args=args,
                replyTo=reply_to
            )

            # FIXME: see `iced` code
            # Aw, i think thats about timeouts
            queue_item.set('channel', reply_to)
            self._sub.subscribe(reply_to)
            receivers_num = self._pub.publish(
                self.service.service_channel_name, msg)
            if receivers_num == 0:
                raise RpcConsummer.EndpointError(
                    rpc_strings['NO_RECEIVER_ON_ENDPOINT_ERROR'].format({'endpoint': self.service.service_channel_name}))
        except:
            self._dequeue_call(id)
            raise
