from .publisher import Publisher
from .subscriber import Subscriber
from .rpc import RpcService
from .message import Message
from .notification import Notification
from .service_info import ServiceInfo
from .lib.threadpool import ThreadPool
from .strings import STRINGS

service_strings = STRINGS['service']
error_strings = STRINGS['errors']

SERVICES_CHANNEL = service_strings['SERVICES_CHANNEL']
PING_MESSAGE = service_strings['PING_MESSAGE']
PONG_MESSAGE = service_strings['PONG_MESSAGE']
INFO_REQUEST_MESSAGE = service_strings['INFO_REQUEST_MESSAGE']
INFO_MESSAGE = service_strings['INFO_MESSAGE']
SETUP_MESSAGE = service_strings['SETUP_MESSAGE']
SHUTDOWN_MESSAGE = service_strings['SHUTDOWN_MESSAGE']
NOT_IMPLEMENTED_MESSAGE = service_strings['NOT_IMPLEMENTED_MESSAGE']

thread_pool = ThreadPool()


class Service:

    channel = SERVICES_CHANNEL

    def __init__(self):
        if not hasattr(self, 'name'):
            raise RuntimeError(service_strings['NAME_PROPERTY_UNDEFINED_ERROR'])

        self._rpc = RpcService(self.name)
        self._pub = Publisher()
        self._sub = Subscriber()

    def get_name(self):
        return self.name

    def get_rpc_instance(self):
        return self._rpc

    def get_info(self):
        raise RuntimeError('Not implemented')

    def _message_handler(self, message):
        thread_pool.add_task(self, '_message_handler_task', message)

    def _message_handler_task(self, message):
        type = message['type']
        if type == NOT_IMPLEMENTED_MESSAGE:
            return
        elif type == PING_MESSAGE:
            msg = Message(type=PONG_MESSAGE, data={'name': self.name})
            self._pub.publish(message['replyTo'], msg)
        elif type == INFO_REQUEST_MESSAGE:
            msg = Message(type=INFO_MESSAGE, data=self.get_info())
            self._pub.publish(message['replyTo'], msg)
        else:
            if not self.on_message(message) and message['replyTo'] != SERVICES_CHANNEL:
                msg = Message(type=NOT_IMPLEMENTED_MESSAGE)
                self._pub.publish(message['replyTo'], msg)

    def on_message(self, message):
        raise RuntimeError(error_strings['METHOD_NOT_IMPLEMENTED_ERROR'].format({'method': 'on_message'}))

    def _notify(self, event):
        msg = Message(type=event, data={'name': self.name})
        self._pub.publish(self.channel, msg)

    def setup(self):
        self._sub.subscribe(self.channel)
        personal_service_channel = self._sub.format_channel_name(self.channel, self.name)
        self._sub.subscribe(personal_service_channel)
        personal_notification_channel = self._sub.format_channel_name(Notification.namespace, self.name)
        self._sub.subscribe(personal_notification_channel)
        self._rpc.setup()
        self._notify(SETUP_MESSAGE)
        self._sub.add_message_handler(self._message_handler)

    def shutdown(self):
        self._sub.destroy()
        self._rpc.shutdown()
        self._notify(SHUTDOWN_MESSAGE)
