import redis
from redis.client import PubSub as RedisPubSub
import threading
import json
from .pubsub import PubSub

class Subscriber(PubSub):

    def __init__(self):
        self._connection = redis.StrictRedis()
        self._pubsub = self._connection.pubsub()
        self._message_handlers = []
        self._polling_active = False
        self._polling_thread = None

    def _message_handler(self, raw_message):
        if raw_message['type'] not in RedisPubSub.PUBLISH_MESSAGE_TYPES:
            return None

        try:
            data = json.loads(raw_message['data'])
            data['channel'] = raw_message['channel']
            if 'replyTo' not in data:
                data['replyTo'] = raw_message['channel']

            for handler in self._message_handlers:
                handler(data)

        except Exception:
            # TODO: call registered error handler or `raise`
            # Save callstack
            raise

    def _message_polling_task(self):
        # Job is async, polling may be not active at this moment
        # if not self._polling_active:
        #     return

        try:
            for msg in self._pubsub.listen():
                self._message_handler(msg)
        except StopIteration:
            # Just exit task
            pass

    def _start_polling_task(self):
        if self._polling_active:
            raise RuntimeError('Polling already active!')
        self._polling_active = True

        t = threading.Thread(target=self._message_polling_task)
        self._polling_thread = t
        t.start()

    def _stop_polling_task(self):
        self._pubsub.unsubscribe()
        if self._polling_thread is not None:
            self._polling_thread.join()
            self._polling_thread = None
        self._polling_active = False

    def add_message_handler(self, fn):
        self._message_handlers.append(fn)

    def subscribe(self, channel_name):
        self._pubsub.subscribe(self.format_channel_name(channel_name))
        if not self._polling_active and len(self._message_handlers) > 0:
            self._start_polling_task()

    def unsubscribe(self, channel_name=None):
        if channel_name is None:
            self._stop_polling_task()
        else:
            self._pubsub.unsubscribe(self.format_channel_name(channel_name))

    def destroy(self):
        self.unsubscribe()
        self._pubsub.close()
